/*
** Module   : AbstractJoin.java
** Abstract : Abstract base class for natural join helpers
**
** Copyright (c) 2006-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ------------------Description-------------------
** 001 ECF 20060309   @25000 Created initial version. Abstract base class for
**                           natural join helper implementations.
** 002 ECF 20060803   @28381 Fixed parameter type defect. Was using the wrong
**                           end of the join for parameter type in a non-
**                           dereference situation.
** 003 SVL 20080421   @38073 Added support for multiple query substitution
**                           parameters.
** 004 CA  20080618   @38866 Save the inverse buffer for later checks.
** 005 ECF 20080822   @39559 Added isServerJoin(). Reports whether join occurs
**                           at database server or within P2J runtime
**                           environment.
** 006 ECF 20090716   @43210 Added RecordBuffer.initialize() calls to c'tor.
** 007 SVL 20120331          Upgraded to Hibernate 4.
*/

package com.goldencode.p2j.persist;

import org.hibernate.type.Type;
import java.io.Serializable;
import java.util.List;

/**
 * The abstract base class for dynamic and server-side join helpers.  Join
 * helpers are used by query implementations to:
 * <ul>
 *   <li>generate HQL where clause snippets which join records using foreign
 *       relations;
 *   <li>provide query substitution parameters if the HQL snippet requires
 *       them at query execution time.
 * </ul>
 * <p>
 * Instances of this class are enclosed within query objects and are used at
 * query prepare time and at query execution time to augment the query with
 * the necessary join semantics.
 *
 * @see  DynamicJoin
 * @see  ServerJoin
 */
abstract class AbstractJoin
{
   /** HQL where clause snippet used to perform the join */
   private final String hql;
   
   /** DMO interface for the inverse (non-local) end of the join */
   private final Class<?> inverseIface;
   
   /** Hibernate types of query substitution parameters (null if none) */
   private final List<Type> parameterTypes;
   
   /** Bufffer to which this component should join via a foreign relation. */
   protected final BufferReference inverse;

   /**
    * Constructor which invokes abstract initialization methods in subclass
    * implementations.
    *
    * @param   local
    *          Local record buffer (the refering end of the join).
    * @param   inverse
    *          Inverse record buffer (the referent end of the join).
    */
   AbstractJoin(RecordBuffer local, RecordBuffer inverse)
   {
      local.initialize();
      inverse.initialize();
      
      // Look up relation information.
      String schema = inverse.getSchema();
      Class<?> localIface = local.getDMOInterface();
      this.inverseIface = inverse.getDMOInterface();
      this.inverse = inverse;
      
      RelationInfo info = DMOIndex.getRelation(schema,
                                               localIface,
                                               inverseIface);
      boolean dereference = (info != null);
      
      if (!dereference)
      {
         info = DMOIndex.getRelation(schema, inverseIface, localIface);
      }
      
      // Assert relation exists.
      if (info == null)
      {
         throw new IllegalArgumentException(
            "No foreign relation between " + localIface.getName() + " and " +
            inverseIface.getName());
      }
      
      // Initialize HQL.
      this.hql = generateHQL(local, inverse, info, dereference);
      
      // Store parameter type.
      this.parameterTypes = getParameterTypes(info, dereference);
      
      // Remaining setup, if any.
      setup(local, inverse, info, dereference);
   }

   /** 
    * Return the inverse buffer used with this join, if any.
    * 
    * @return  The inverse buffer used with this join.
    */
   public BufferReference getInverse()
   {
      return inverse;
   }
   
   /**
    * Report whether the join represented by this object takes place at the
    * database server, or within runtime code.
    * 
    * @return  <code>true</code> if the join happens at the server;
    *          <code>false</code> if it happens in the runtime.
    */
   protected abstract boolean isServerJoin();
   
   /**
    * Generate the HQL where clause snippet which will be inserted into the
    * overall HQL statement executed by the enclosing query.
    *
    * @param   local
    *          Local record buffer (the refering end of the join).
    * @param   inverse
    *          Inverse record buffer (the referent end of the join).
    * @param   info
    *          Relation descriptor object.
    * @param   dereference
    *          <code>true</code> if the local DMO contains a reference to the
    *          foreign record, which must be dereferenced in the HQL query;
    *          <code>false</code> if the local DMO <i>is</i> the foreign
    *          record, referenced by the inverse DMO.
    *
    * @return  HQL where clause snippet which expresses the join.
    */
   protected abstract String generateHQL(RecordBuffer local,
                                         RecordBuffer inverse,
                                         RelationInfo info,
                                         boolean dereference);

   /**
    * Determine the Hibernate type of the query substitution parameters, if
    * any.  The parameter(s) generally will be a Hibernate entity type which
    * wrappers the implementation class of the inverse DMO or the types of
    * DMO fields.
    *
    * @param   info
    *          Relation descriptor object.
    * @param   dereference
    *          <code>true</code> if the local DMO contains a reference to the
    *          foreign record, which must be dereferenced in the HQL query;
    *          <code>false</code> if the local DMO <i>is</i> the foreign
    *          record, referenced by the inverse DMO.
    *
    * @return  The list of appropriate Hibernate types, or <code>null</code>
    *          if no query substitution parameters are used for this join.
    */
   protected abstract List<Type> getParameterTypes(RelationInfo info,
                                                   boolean dereference);

   /**
    * Perform any additional setup necessary for a particular join helper
    * implementation.  This base implementation does nothing.
    *
    * @param   local
    *          Local record buffer (the refering end of the join).
    * @param   inverse
    *          Inverse record buffer (the referent end of the join).
    * @param   info
    *          Relation descriptor object.
    * @param   dereference
    *          <code>true</code> if the local DMO contains a reference to the
    *          foreign record, which must be dereferenced in the HQL query;
    *          <code>false</code> if the local DMO <i>is</i> the foreign
    *          record, referenced by the inverse DMO.
    */
   protected void setup(RecordBuffer local,
                        RecordBuffer inverse,
                        RelationInfo info,
                        boolean dereference)
   {
   }
   
   /**
    * Get the values of DMO instance fields or the DMO instance itself which
    * will be used as the query substitution parameter(s). This base
    * implementation returns <code>null</code>. Subclasses which need a
    * parameter must override this method.
    *
    * @return  <code>null</code>.
    */
   List<Serializable> getParameters()
   {
      return null;
   }
   
   /**
    * Get the DMO interface for the inverse (non-local) end of the join.
    *
    * @return  Inverse DMO interface.
    */
   Class<?> getInverseInterface()
   {
      return inverseIface;
   }
   
   /**
    * Get the HQL where clause snippet used to perform the join.  The
    * enclosing query embeds this phrase into the overall HQL statement
    * before executing the query.
    *
    * @return  HQL where clause snippet used to perform the join.
    */
   String getHQL()
   {
      return hql;
   }
   
   /**
    * Get the type of the query substitution parameters, if any, which was
    * determined at construction.
    *
    * @return  The list of appropriate Hibernate types, or <code>null</code>
    *          if no query substitution parameters are used for this join.
    */
   List<Type> getParameterTypes()
   {
      return parameterTypes;
   }
}
