/*
** Module   : DynamicJoin.java
** Abstract : Natural join helper implementation for dynamic queries
**
** Copyright (c) 2006-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 ECF 20060309   @25002 Created initial version. Natural join helper
**                           implementation used by dynamic query types.
** 002 ECF 20060621   @27516 Fix to setup method. Use lowercased short
**                           name of DMO interface when composing name of
**                           foreign record property.
** 003 ECF 20060803   @28389 Minor change to getParameterType() method.
**                           Renamed parameter and changed javadoc.
** 004 ECF 20070629   @35297 Minor optimization. Replaced StringBuffer
**                           with StringBuilder.
** 005 SVL 20080421   @38075 Updated to meet the AbstractJoin function 
**                           signatures which may return multiple query
**                           substitution parameters.
** 006 CA  20080515   @38283 Fixed regression introduced by H005: the
**                           parameter type must always be retrieved
**                           from the foreign class.
** 007 ECF 20080822   @39560 Added isServerJoin(). Reports whether join
**                           occurs at database server or within P2J
**                           runtime environment.
** 008 GES 20090424   @41938 Import change.
** 009 ECF 20090428   @42973 Fixed setup(). Use record buffer's snapshot to
**                           resolve parameter in non-dereference case.
** 010 SVL 20120331          Upgraded to Hibernate 4.
*/

package com.goldencode.p2j.persist;

import org.hibernate.*;
import org.hibernate.type.Type;
import java.util.List;
import java.util.Collections;
import java.io.Serializable;
import com.goldencode.util.*;

/**
 * An implementation of a natural join helper used by dynamic query types.
 * The HQL query string snippet created by this class joins the local DMO
 * with the inverse DMO using a query substitution parameter (<code>?</code>)
 * to represent the shared record.  For instance, given the three table
 * relationship:
 * <pre>
 *    Person [one-to-many] ---> PersonAddress [many-to-one] ---> Address
 * </pre>
 * and the following multi-table join in Progress
 * <pre>
 *    for each person,
 *        each person-address of person,
 *        each address of person-address:
 *       ...
 * </pre>
 * the first join (<code>person-address of person</code>) is represented as
 * follows:
 * <pre>
 *    personAddress.personRecord = ?
 * </pre>
 * where <code>PersonAddress</code> is the local DMO and Person the foreign
 * DMO;  and the second join (<code>address of person-address</code>) is
 * represented as follows:
 * <pre>
 *    address = ?
 * </pre>
 * where <code>Address</code> is the local DMO and PersonAddress the foreign
 * DMO.
 */
final class DynamicJoin
extends AbstractJoin
{
   /** An object which resolves the substitution parameter */
   private ParameterResolver resolver;
   
   /**
    * Constructor.
    *
    * @param   local
    *          Local record buffer (the refering end of the join).
    * @param   inverse
    *          Inverse record buffer (the referent end of the join).
    */
   DynamicJoin(RecordBuffer local, final RecordBuffer inverse)
   {
      super(local, inverse);
   }
   
   /**
    * Report whether the join represented by this object takes place at the
    * database server, or within runtime code.
    * 
    * @return  <code>true</code> if the join happens at the server;
    *          <code>false</code> if it happens in the runtime.
    */
   protected boolean isServerJoin()
   {
      return false;
   }
   
   /**
    * Generate the HQL where clause snippet which will be inserted into the
    * overall HQL statement executed by the enclosing query.  See the class
    * description for additional details on the composition of the clause.
    *
    * @param   local
    *          Local record buffer (the refering end of the join).
    * @param   inverse
    *          Inverse record buffer (the referent end of the join).
    * @param   info
    *          Relation descriptor object.
    * @param   dereference
    *          <code>true</code> if the local DMO contains a reference to the
    *          foreign record, which must be dereferenced in the HQL query;
    *          <code>false</code> if the local DMO <i>is</i> the foreign
    *          record, referenced by the inverse DMO.
    *
    * @return  HQL where clause snippet which expresses the join.
    */
   protected String generateHQL(RecordBuffer local,
                                RecordBuffer inverse,
                                RelationInfo info,
                                boolean dereference)
   {
      StringBuilder buf = new StringBuilder(local.getDMOAlias());
      if (dereference)
      {
         buf.append(".");
         buf.append(info.getForeignAlias());
         buf.append("Record");
      }
      buf.append(" = ?");
      
      return buf.toString();
   }
   
   /**
    * Determine the Hibernate type of the query substitution parameter.  This
    * will be a Hibernate entity type which wrappers the implementation class
    * of the buffer.
    *
    * @param   info
    *          Relation descriptor object.
    * @param   dereference
    *          <code>true</code> if the local DMO contains a reference to the
    *          foreign record, which must be dereferenced in the HQL query;
    *          <code>false</code> if the local DMO <i>is</i> the foreign
    *          record, referenced by the inverse DMO.
    *
    * @return  A singleton list with the appropriate Hibernate type for the
    *          query substitution parameter is used for this join.
    */
   protected List<Type> getParameterTypes(RelationInfo info,
                                          boolean dereference)
   {
      // always return the foreign class, because the RelationInfo object is
      // already dereferenced by the AbstractJoin c'tor
      TypeHelper th = inverse.buffer().getPersistence().getSession().getTypeHelper();
      return Collections.singletonList(th.entity(info.getForeignClass()));
   }
   
   /**
    * Perform additional setup necessary to create the object which will
    * resolve the query substitution parameter at query execution time.
    *
    * @param   local
    *          Local record buffer (the refering end of the join).
    * @param   inverse
    *          Inverse record buffer (the referent end of the join).
    * @param   info
    *          Relation descriptor object.
    * @param   dereference
    *          <code>true</code> if the local DMO contains a reference to the
    *          foreign record, which must be dereferenced in the HQL query;
    *          <code>false</code> if the local DMO <i>is</i> the foreign
    *          record, referenced by the inverse DMO.
    */
   protected void setup(RecordBuffer local,
                        final RecordBuffer inverse,
                        RelationInfo info,
                        boolean dereference)
   {
      // Initialize query substitution parameter resolver.
      if (dereference)
      {
         this.resolver = new ParameterResolver()
         {
            public List<Serializable> resolve()
            {
               return Collections.singletonList(
                  (Serializable) inverse.getCurrentRecord());
            }
         };
      }
      else
      {
         StringBuilder buf = new StringBuilder();
         buf.append(StringHelper.changeCase(local.getDMOName(), false));
         buf.append("Record");
         final String prop = buf.toString();
         
         this.resolver = new ParameterResolver()
         {
            private FieldReference ref = new FieldReference(inverse, prop);
            
            public List<Serializable> resolve()
            {
               return Collections.singletonList(
                  (Serializable) ref.get(inverse.getSnapshot()));
            }
         };
      }
   }
   
   /**
    * Get the DMO instance which will be used as the query substitution
    * parameter.  Utilizes the inner helper class {@link
    * DynamicJoin.ParameterResolver} to do its work.
    *
    * @return  Query substitution parameter;  a singleton list with DMO
    *          instance.
    */
   List<Serializable> getParameters()
   {
      return resolver.resolve();
   }
   
   /**
    * An internal API used by this class to abstract the resolution of the
    * query substitution parameter at query execution time.
    */
   private interface ParameterResolver
   {
      /**
       * Get the DMO instance which will be used as the query substitution
       * parameter.
       *
       * @return  Query substitution parameter;  a singleton list with DMO
       *          instance.
       */
      public List<Serializable> resolve();
   }
}
