/*
** Module   : TableHighIdentityManager.java
** Abstract : Identity manager implementation which uses high-water ID of each
**            table as a starting point for new IDs
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SVL 20071231   @36636 Created initial version. Default
**                           implementation of IdentityManager interface.
** 002 ECF 20080805   @39305 Adapted from DefaultIdentityManager. Modified
**                           to produce 64-bit keys. Conformed to new
**                           IdentityManager API.
** 003 SVL 20080901   @39646 Implements IdentityPoolManager instead of
**                           IdentityManager.
** 004 SVL 20120331          Upgraded to Hibernate 4.
*/

package com.goldencode.p2j.persist.id; 

import java.io.Serializable;
import java.util.*;
import org.hibernate.*;
import org.hibernate.cfg.*;
import org.hibernate.mapping.RootClass;
import org.apache.commons.logging.*;
import com.goldencode.p2j.persist.*;

/**
 * @deprecated Use {@link SequenceIdentityManager} or other identity manager
 *             which supports database-wide unique ids.
 *
 * Implementation of the {@link IdentityManager} interface.  During server
 * initialization, each database table is queried to determine its maximum
 * primary key ID.  This information is stored in a map for later use in
 * providing the next available ID for each table.  The per-table value is
 * incremented for each {@link #nextPrimaryKey(String) ID request}.
 * <p>
 * Current limitations:
 * <ul>
 *   <li><b>Does not produce database-wide unique IDs</b>.  Note that this
 *       does not match Progress behavior.
 *   <li>Wrapping to a negative number after we increment beyond
 *       <code>Long.MAX_VALUE</code> is not handled at this time.
 *   <li>Does not recycle IDs which are allocated but never used.
 *   <li>Does not reclaim IDs for deleted records.
 *   <li>Does not detect/repair fragmentation of the primary key address
 *       space for a table;  existing gaps in primary keys are never plugged.
 *       Note that a solution for this item may preclude having to address the
 *       previous three items.
 * </ul>
 */
public final class TableHighIdentityManager
implements IdentityPoolManager
{
   /** Logger */
   private static final Log LOG = LogFactory.getLog(IdentityManager.class);
   
   /** Map for storing table name -> current maximum ID relations. */
   private final Map<String, Long> ids;
   
   /** Persistence service object used to interact with the database */
   private Persistence persistence = null;

   /**
    * Default constructor.
    */
   public TableHighIdentityManager()
   {
      ids = new HashMap<String, Long>();
   }

   /**
    * Specify the database for the identity manager.
    * <code>TableHighIdentityManager</code> is initialized here - current
    * maximum IDs for each table are retrieved.
    *
    * @param   persistence
    *          Persistence service object associated with identity manager.
    */
   public void setPersistence(Persistence persistence)
   {
      this.persistence = persistence;
      Database database = persistence.getDatabase();
      
      Session session = null;
      
      boolean debug = LOG.isDebugEnabled();

      try
      {
         SessionFactory factory = DatabaseManager.getSessionFactory(database);
         Configuration cfg =
            DatabaseManager.getHibernateConfiguration(database);
         Iterator iter = cfg.getClassMappings();
         session = factory.openSession();
         
         while (iter.hasNext())
         {
            Object obj = iter.next();
            if (obj instanceof RootClass)
            {
               RootClass rc = (RootClass) obj;
               String className = rc.getClassName();
               String tableName = rc.getTable().getName();
               StringBuilder buf = new StringBuilder();
               buf.append("select max(obj.id) from ");
               buf.append(className);
               buf.append(" obj");
               Query query = session.createQuery(buf.toString());

               Object res = query.uniqueResult();
               Long max = (res == null ? 0 : (Long) res);
               synchronized (ids)
               {
                  ids.put(tableName, max);
               }
               
               if (debug)
               {
                  buf = new StringBuilder();
                  buf.append("Last ID for '");
                  buf.append(database.getName());
                  buf.append(".");
                  buf.append(tableName);
                  buf.append("':  ");
                  buf.append(max);
                  LOG.debug(buf.toString());
               }
            }
         }
      }
      catch (Exception exc)
      {
         String msg = "Error initializing identity manager for database '" +
                      database.getName() +
                      "'";
         LOG.error(msg, exc);
         
         throw new RuntimeException(msg, exc);
      }
      finally
      {
         if (session != null)
         {
            session.close();
         }
      }
   }

   /**
    * Return the next primary key for the specified table.
    *
    * @param   table
    *          Table for which the next primary key will be returned.
    *
    * @return  The next primary key for the given table.
    * 
    * @throws  PersistenceException
    *          if there is an error determining the next primary key ID.
    */
   public Serializable nextPrimaryKey(String table)
   throws PersistenceException
   {
      synchronized (ids)
      {
         Long id = ids.get(table);
         if (id == null)
         {
            throw new PersistenceException(
               "Cannot locate next ID for table '" +
               persistence.getDatabase().getName() +
               "." +
               table +
               "'");
         }
         
         id++;
         ids.put(table, id);
         
         return id;
      }
   }

   /**
    * Inform ithe dentity manager that the given primary keys can be reused.
    * No-op in this implementation.
    *
    * @param keys
    *        Keys for reuse.
    */
   public void reclaimKeys(Serializable[] keys)
   {
   }

   /**
    * Sets identity pool which can be used by this identity manager. No-op in
    * this implementaion.
    *
    * @param identityPool
    *        Identity pool for use.
    */
   public void setIdentityPool(IdentityPool identityPool)
   {
   }
}
