/*
** Module   : H2SequenceHandler.java
** Abstract : Static class that handles sequence functions and statements 
**            from Progress / OpenEdge 10.x.
**
** Copyright (c) 2012-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------Description---------------------
** 001 OM  20121116 Created initial version with CURRVAL, NEXTVAL, and 
**                  SETVAL support only.
** 002 SVL 20120331 Upgraded to Hibernate 4.
*/

package com.goldencode.p2j.persist.sequence;

import java.util.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.p2j.util.*;
import org.hibernate.*;

/**
 * H2 implementation of sequences using a hybrid native SQL sequence
 * primitives and 4GL fixes backed by directory.
 * The primary goal is to detect unknown (?) values that are not supported by
 * H2 and also full management of cycling sequences.
 * H2 only support incremental sequences, no cycling, no limits.
 */
public class H2SequenceHandler
extends SequenceHandler
{
   /**
    * The list of sequences grouped by databases.
    * It uses lazy initialization.
    */
   private HashMap<String, LinkedList<String>> sequences = null;

   /**
    * Queries the current value of a sequence.
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. 
    *          It is assumed it is not null and valid.
    *
    * @return the current value of the sequence or unknown value (?) if
    *          sequence/database not found or the sequence is not cycling
    *          and has already passed the maximum value.
    */
   @Override
   public int64 getCurrentValue(String seqName, String ldbName)
   {
      Sequence seq = SequenceManager.getSequence(ldbName, seqName);

      Persistence persistence = ConnectionManager.getPersistence(ldbName);
      P2JDialect dialect = (P2JDialect) persistence.getDialect();
      String strQuery = dialect.getSequenceCurrValString(seqName);
      long ret = 0;

      try
      {
         ScrollableResults rs = persistence.executeSQLQuery(strQuery);
         if (rs.next())
         {
            ret = (Long) rs.get(0);
         }
      }
      catch (PersistenceException pe)
      {
         pe.printStackTrace();
         return new int64(); // return unknown ?
      }

      // check the result against the dmo-index properties of the sequence
      long inc = seq.getIncrement();
      if (inc > 0)
      {
         // ascendant sequence
         if (ret > seq.getMax())
         {
            // current-value return the last valid value
            ret = seq.getMax();
         }
      }
      else if (inc < 0)
      {
         // descendant sequence
         if (ret < seq.getMin())
         {
            // current-value return the last valid value
            ret = seq.getMin();
         }
      }

      return new int64(ret);
   }

   /**
    * Computes and returns the next value of a sequence.
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. 
    *          It is assumed it is not null and valid.
    *
    * @return the current value of the sequence or unknown value (?) if
    *          sequence/database not found or the sequence is not cycling
    *          and has passed the maximum/minimum value.
    */
   @Override
   public int64 getNextValue(String seqName, String ldbName)
   {
      Sequence seq = SequenceManager.getSequence(ldbName, seqName);

      Persistence persistence = ConnectionManager.getPersistence(ldbName);
      P2JDialect dialect = (P2JDialect) persistence.getDialect();
      String strQuery = dialect.getSequenceNextValString(seqName);
      long ret = 0;

      try
      {
         ScrollableResults rs = persistence.executeSQLQuery(strQuery);
         if (rs.next())
         {
            ret = (Long) rs.get(0);
         }
      }
      catch (PersistenceException pe)
      {
         pe.printStackTrace();
         return new int64(); // return unknown ?
      }

      // check the result against the dmo-index properties of the sequence
      long inc = seq.getIncrement();
      if (inc > 0)
      {
         // ascendant sequence
         if (ret > seq.getMax())
         {
            if (seq.isCycle())
            {
               // compute the correct value by warping downwards
               ret = seq.getMin();

               // also set to database:
               setValue(seqName, ldbName, ret);
            }
            else
            {
               // terminating sequence, returning unknown
               return new int64();
            }
         }
      }
      else if (inc < 0)
      {
         // descendant sequence
         if (ret < seq.getMin())
         {
            if (seq.isCycle())
            {
               // compute the correct value by warping upwards
               ret = seq.getMax();

               // also set to database:
               setValue(seqName, ldbName, ret);
            }
            else
            {
               // terminating sequence, returning unknown
               return new int64();
            }
         }
      }

      return new int64(ret);
   }

   /**
    * Initialize (reset) the current value of a sequence.
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. 
    *          It is assumed it is not null and valid.
    * @param   newVal
    *          The value the sequence will be reseted to.
    */
   @Override
   public void setValue(String seqName, String ldbName, long newVal)
   {
      Sequence seq = SequenceManager.getSequence(ldbName, seqName);

      /* H2 database server sets the current value of a sequence so as 
       * the following <code>next value</code> is called this newVal argument
       * is returned so the actual current value is newVal - increment. 
       * Postgres 4GL instead sets the value and it is accessible via 
       * <code>(dynamic-)current-value</code> while 
       * <code>(dynamic-)next-value</code> will return newVal + increment.
       * The solution is to set the value already incremented into database.*/
      long newValFixed = newVal + seq.getIncrement();

      Persistence persistence = ConnectionManager.getPersistence(ldbName);
      P2JDialect dialect = (P2JDialect) persistence.getDialect();
      String strQuery = dialect.getSequenceSetValString(seqName,newValFixed);

      try
      {
         persistence.executeSQL(strQuery);
      }
      catch (PersistenceException pe)
      {
         pe.printStackTrace();
      }
   }

   /**
    * Test whether the database contains a sequence. This is useful to
    * identify the correct database/sequences when a 4GL static function /
    * statement is called without the database parameter. 
    *
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. If is null then
    *          the first connected database is assumed.
    *
    * @return  true if the sequence is defined in the database
    */
   @Override
   public boolean contains(String ldbName, String seqName)
   {
      // lazy initialization
      if (sequences == null)
      {
         sequences = new HashMap<String, LinkedList<String>>();
      }
         
      // make sure the structure is initialized for this database:
      LinkedList<String> crtSeqList = null;
      if (!sequences.containsKey(ldbName))
      {
         // create the list 
         crtSeqList = new LinkedList<String>();
         sequences.put(ldbName, crtSeqList);

         Persistence persistence = ConnectionManager.getPersistence(ldbName);
         try
         {
            ScrollableResults rs = persistence.executeSQLQuery(
                        "SELECT SEQUENCE_NAME " +
                        "FROM INFORMATION_SCHEMA.SEQUENCES;");
            while (rs.next())
            {
               crtSeqList.add((String) rs.get(0));
            }
         }
         catch (PersistenceException pe)
         {
            pe.printStackTrace();
            return false;
         }
      }
      else
      {
         crtSeqList = sequences.get(ldbName);
      }

      // now check whether the sequence is in database
      return crtSeqList.contains(seqName);
   }
}
