/*
** Module   : LogicalUserType.java
** Abstract : Custom user type to wrap P2J logical for Hibernate access
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 ECF 20050630   @21592 Created initial version. Maps P2J logical to
**                           SQL bit.
** 002 ECF 20050729   @21956 Refactored to move nullSafeSet method
**                           implementation up to superclass. This was
**                           necessary to add centralized logging for
**                           prepared statement parameter binding.
** 003 ECF 20050809   @22076 Fixed nullSafeGet and added logging. Defect
**                           was that method was checking whether the
**                           result set column was null *before* reading
**                           data at that column. Also made class final.
** 004 ECF 20080806   @39299 Moved to type sub-package of persist.
** 005 SVL 20120331          Upgraded to Hibernate 4.
*/

package com.goldencode.p2j.persist.type;

import java.sql.*;
import com.goldencode.p2j.util.*;
import org.hibernate.engine.spi.*;
import org.hibernate.usertype.*;

/**
 * Implementation of a custom, Hibernate user type for the {@link
 * com.goldencode.p2j.util.logical P2J logical}. This permits Hibernate to
 * convert this data type to and from the SQL BIT data type. The unknown
 * value is mapped to SQL NULL.
 */
public final class LogicalUserType
extends AbstractWrapperUserType
implements EnhancedUserType
{
   /** Single element array containing the SQL type constant being mapped */
   private static final int[] SQL_TYPES = { Types.BIT };
   
   /**
    * Get the array of SQL data type constants (in this case only a single
    * constant), which this classes uses to map a logical to/from a database
    * record.
    *
    * @return  Array containing a single type constant:  BIT.
    */
   public int[] sqlTypes()
   {
      return SQL_TYPES;
   }
   
   /**
    * Gets the Java data type class mapped by this custom type.
    *
    * @return  The {@link com.goldencode.p2j.util.logical P2J logical} class.
    */
   public Class returnedClass()
   {
      return logical.class;
   }
   
   /**
    * Create and return a deep copy of the given <code>logical</code>
    * instance.
    *
    * @param   value
    *          The <code>logical</code> instance to be copied.
    *
    * @return  A new instance of <code>logical</code> which contains the
    *          same internal value as <code>value</code>.
    */
   public Object deepCopy(Object value)
   {
      return new logical((logical) value);
   }
   
   /**
    * Create an instance of the data type mapped by this class, based upon
    * a data value in the specified result set. If the corresponding data
    * value is NULL in the result set, return a <code>logical</code>
    * instance which is initialized to the unknown value.
    *
    * @param   rs
    *          Result set positioned at the record from which we will access
    *          a <code>boolean</code>.
    * @param   names
    *          Array of column names. In this case, it will contain only one
    *          element, which represents the name of the column from which
    *          we will access our data.
    * @param   session
    *          Not used.
    * @param   owner
    *          Owning entity; not used.
    *
    * @return  A <code>logical</code> instance, possibly initialized to the
    *          unknown value.
    *
    * @throws  SQLException
    *          if an error occurs accessing the result set.
    */
   public Object nullSafeGet(ResultSet rs,
                             String[] names,
                             SessionImplementor session,
                             Object owner)
   throws SQLException
   {
      boolean value = rs.getBoolean(names[0]);
      
      if (rs.wasNull())
      {
         if (LOG_TRACE)
         {
            LOG.trace("returning unknown value as column: " + names[0]);
         }
         return new logical();
      }
      
      logical wrapper = new logical(value);
      if (LOG_TRACE)
      {
         LOG.trace("returning '" + toString(wrapper) + "' as column: " +
                   names[0]);
      }
      
      return wrapper;
   }
   
   /**
    * Set a <code>boolean</code> value into the given prepared statement,
    * based upon the <code>logical</code> instance <code>value</code>.
    *
    * @param   ps
    *          Prepared statement into which we will set our data.
    * @param   value
    *          An <code>logical</code> instance.
    * @param   index
    *          Index in the statement at which to write the data.
    *
    * @throws  SQLException
    *          if an error occurs setting data into the statement.
    */
   protected void bindNonNull(PreparedStatement ps,
                              BaseDataType value,
                              int index)
   throws SQLException
   {
      ps.setBoolean(index, ((logical) value).getValue());
   }

   /**
	 * Return an SQL literal representation of the value.
    *
    * @param  value
    *         Value which representation we want to get.
    *
    * @return SQL literal representation of the value.
	 */
   public String objectToSQLString(Object value)
   {
      return value.toString();
   }

   /**
	 * Return a string representation of this value, as it should appear in an XML document.
    *
    * @param  value
    *         Value to covert to an XML representation.
    *
    * @return <code>null</code>
	 */
   public String toXMLString(Object value)
   {
      return null;
   }

   /**
	 * Parse a string representation of this value, as it appears in an XML document.
    *
    * @param  xmlValue
    *         XML representation of the value.
    *
    * @return <code>null</code>
    */
   public Object fromXMLString(String xmlValue)
   {
      return null;
   }
}
